from enum import Enum


class CameraType(str, Enum):
    PINHOLE = "pinhole"
    FISHEYE = "fisheye"
    KANNALA = "kannala"


class InvalidatedReasonInput(str, Enum):
    BAD_CONTENT = "bad-content"
    SLAM_RECORRECTION = "slam-rerun"
    DUPLICATE = "duplicate"
    INCORRECTLY_CREATED = "incorrectly-created"


class InputBatchStatus(str, Enum):
    CREATED = 'created'
    OPEN = 'open'
    READY = 'ready'
    INPROGESS = 'in-progress'
    COMPLETED = 'completed'


class InputStatus(str, Enum):
    PROCESSING = "processing"
    CREATED = "created"
    FAILED = "failed"
    INVALIDATED_BROKEN_INPUT = "invalidated:broken-input"
    INVALIDATED_SLAM_RECORRECTION = "invalidated:slam-rerun"
    INVALIDATED_DUPLICATE = "invalidated:duplicate"
    INVALIDATED_INCORRECTLY_CREATED = "invalidated:incorrectly-created"



"""Utility functions for Input API """

import mimetypes
from collections.abc import Mapping
from datetime import datetime
from pathlib import Path
from PIL import Image as PILImage
import dateutil.parser
from urllib3.util import Url, parse_url


GCS_SCHEME = "gs"


def ts_to_dt(date_string: str) -> datetime:
    """
    Parse string datetime into datetime
    """
    return dateutil.parser.parse(date_string)


def filter_none(js: dict) -> dict:
    if isinstance(js, Mapping):
        return {k: filter_none(v) for k, v in js.items() if v is not None}
    else:
        return js


def get_content_type(filename: str) -> str:
    # https://developer.mozilla.org/en-US/docs/Web/HTTP/Basics_of_HTTP/MIME_types/Complete_list_of_MIME_types
    if filename.split(".")[-1] == "csv":
        content_type = "text/csv"
    else:
        content_type = mimetypes.guess_type(filename)[0]
        content_type = content_type if content_type is not None else 'application/octet-stream'
    return content_type


def get_resource_id(signed_url: str) -> str:
    url = parse_url(signed_url)
    resource_id = Url(scheme=GCS_SCHEME, path=url.path)
    return str(resource_id).replace("///", "//")


def get_image_dimensions(image_path: str) -> dict:
    fi = Path(image_path).expanduser()
    with PILImage.open(fi) as im:
        width, height = im.size
        return {"width": width, "height": height}


import logging
from typing import List, Optional

from .abstract_models import SensorCalibration
from .calibration import CameraProperty
from .enums import CameraType

log = logging.getLogger(__name__)


class CameraCalibration(SensorCalibration):
    def __init__(self, position: List[float], rotation_quaternion: List[float],
                 camera_matrix: List[float], camera_properties: CameraProperty,
                 distortion_coefficients: List[float], image_height: int, image_width: int,
                 undistortion_coefficients: Optional[List[float]]):

        self.position = position
        self.rotation_quaternion = rotation_quaternion

        self.camera_matrix = camera_matrix
        self.camera_properties = camera_properties
        self.distortion_coefficients = distortion_coefficients
        self.image_height = image_height
        self.image_width = image_width
        self.undistortion_coefficients = undistortion_coefficients

        assert(len(position) == 3)
        assert(len(rotation_quaternion) == 4)
        assert(len(camera_matrix) == 9)

        if camera_properties.camera_type == CameraType.KANNALA:
            assert(undistortion_coefficients is not None)
            assert(len(distortion_coefficients) == 4 and len(undistortion_coefficients) == 4)
        else:
            assert(len(distortion_coefficients) == 5)

        log.warning(f"DEPRECATED: Use the new, explicit camera calibration model CameraCalibrationExplicit.")  # noqa:E501

    def to_dict(self):
        base = {
            "position": self.position,
            "rotation_quaternion": self.rotation_quaternion,
            "camera_matrix": self.camera_matrix,
            "camera_properties": self.camera_properties.to_dict(),
            "distortion_coefficients": self.distortion_coefficients,
            "image_height": self.image_height,
            "image_width": self.image_width
        }

        if self.undistortion_coefficients is not None:
            base["undistortion_coefficients"] = self.undistortion_coefficients

        return base


class LidarCalibration(SensorCalibration):
    def __init__(self, position: List[float], rotation_quaternion: List[float]):
        self.position = position
        self.rotation_quaternion = rotation_quaternion

        assert(len(position) == 3)
        assert(len(rotation_quaternion) == 4)
        log.warning(f"DEPRECATED: Use the new, explicit lidar calibration model LidarCalibrationExplicit.")  # noqa:E501

    def to_dict(self):
        return {
            "position": self.position,
            "rotation_quaternion": self.rotation_quaternion
        }



from typing import Dict


class RequestCall:
    def to_dict(self) -> Dict:
        raise NotImplementedError


class SensorCalibration(RequestCall):

    def to_dict(self) -> Dict:
        raise NotImplementedError


class Response:
    @staticmethod
    def from_json(js: dict):
        raise NotImplementedError



"""API MODEL."""
from typing import List, Mapping, Dict, Optional, Union, Any
from datetime import datetime
import dateutil.parser

from .model.abstract_models import RequestCall, Response
from .model.enums import InputBatchStatus, InvalidatedReasonInput, InputStatus
# To preserve old import statements
from .model.enums import CameraType
from .model.calibration import CameraProperty
from .model.calibration_legacy import CameraCalibration, LidarCalibration
from .model.calibration import CameraCalibrationExplicit, LidarCalibrationExplicit
ENVELOPED_JSON_TAG = "data"


def ts_to_dt(date_string: str) -> datetime:
    return dateutil.parser.parse(date_string)


#
# Request Calls
#


class Image(RequestCall):
    def __init__(self, filename: str,
                 width: Optional[int] = None,
                 height: Optional[int] = None,
                 source: str = "CAM"):

        self.filename = filename
        self.width = width
        self.height = height
        self.source = source

    def to_dict(self) -> dict:
        return dict(filename=self.filename,
                    width=self.width,
                    height=self.height,
                    source=self.source)


class ImagesFiles(RequestCall):
    def __init__(self, images: List[Image]):
        self.images = images

    def to_dict(self) -> dict:
        return dict(images=[image.to_dict() for image in self.images])


class Video(RequestCall):
    def __init__(self, filename: str,
                 width: int,
                 height: int,
                 source: str = "CAM"):

        self.filename = filename
        self.width = width
        self.height = height
        self.source = source

    def to_dict(self) -> dict:
        return dict(filename=self.filename,
                    width=self.width,
                    height=self.height,
                    source=self.source)


class PointCloud(RequestCall):
    def __init__(self, filename: str, source: Optional[str] = "lidar"):
        self.filename = filename
        self.source = source

    def to_dict(self) -> dict:
        return dict(filename=self.filename,
                    source=self.source)


class PointCloudsWithImages(RequestCall):
    def __init__(self, images: List[Image], point_clouds: List[PointCloud]):
        self.images = images
        self.point_clouds = point_clouds

    def to_dict(self):
        return dict(images=[image.to_dict() for image in self.images],
                    pointClouds=[pc.to_dict() for pc in self.point_clouds])


class PointCloudFiles(RequestCall):
    def __init__(self, point_clouds: List[PointCloud]):
        self.point_clouds = point_clouds

    def to_dict(self) -> dict:
        return dict(pointClouds=[point_cloud.to_dict() for point_cloud in self.point_clouds])


class Calibration(RequestCall):
    def __init__(self, calibration_dict: Dict[str, Union[CameraCalibration, CameraCalibrationExplicit, LidarCalibrationExplicit, LidarCalibration]]):  # noqa:E501
        self.calibration_dict = calibration_dict

    def to_dict(self):
        return dict(
            [(k, v.to_dict()) for (k, v) in self.calibration_dict.items()]
        )


class CalibrationSpec(RequestCall):
    def __init__(self, external_id: str,
                 calibration: Calibration):
        self.external_id = external_id
        self.calibration = calibration

    def to_dict(self):
        return {
            'externalId': self.external_id,
            'calibration': self.calibration.to_dict()
        }


class SourceSpecification(RequestCall):
    def __init__(self, source_to_pretty_name: Optional[Dict[str, str]] = None,
                 source_order: Optional[List[str]] = None):

        self.source_to_pretty_name = source_to_pretty_name
        self.source_order = source_order

    def to_dict(self):
        as_dict = {}
        if self.source_to_pretty_name is not None:
            as_dict['sourceToPrettyName'] = self.source_to_pretty_name
        if self.source_order is not None:
            as_dict['sourceOrder'] = self.source_order

        return as_dict


class SceneMetaData(RequestCall):
    def __init__(
        self, external_id: str, source_specification: Optional[SourceSpecification] = None
    ):
        self.external_id = external_id
        self.source_specification = source_specification

    def to_dict(self):
        as_dict = dict(externalId=self.external_id)
        if self.source_specification is not None:
            as_dict["sourceSpecification"] = self.source_specification.to_dict()
        return as_dict


class CalibratedSceneMetaData(SceneMetaData):
    def __init__(self, external_id: str,
                 calibration_id: int,
                 source_specification: Optional[SourceSpecification] = None):

        super().__init__(external_id, source_specification)
        self.calibration_id = calibration_id

    def to_dict(self):
        as_dict = super().to_dict()
        if self.calibration_id is not None:
            as_dict['calibrationId'] = self.calibration_id

        return as_dict


class TimeCalibration(RequestCall):
    def __init__(self, offset_spec: Dict[str, float]):
        self.offset_spec = offset_spec

    def to_dict(self) -> dict:
        return dict(offsetSpec=self.offset_spec)


class TimeSpecification(RequestCall):
    def __init__(self, start_ts: float, end_ts: float, time_calibration: TimeCalibration):
        self.start_ts = start_ts
        self.end_ts = end_ts
        self.time_calibration = time_calibration

    def to_dict(self) -> dict:

        as_dict: Dict[Any, Any] = dict(timeCalibration=self.time_calibration.to_dict())

        if self.start_ts:
            as_dict["startTs"] = self.start_ts
        if self.end_ts:
            as_dict["endTs"] = self.end_ts

        return as_dict


class SlamDynamicObjects():
    def __init__(self,
                 cuboids: List[str],
                 cuboid_timespans: List[str]
                 ):
        self.cuboids = cuboids
        self.cuboid_timespans = cuboid_timespans

    def to_dict(self):
        return dict(cuboids=self.cuboids, cuboidTimespans=self.cuboid_timespans)


class SlamMetaData(CalibratedSceneMetaData):
    def __init__(self, external_id: str,
                 vehicle_data: List[str],
                 dynamic_objects: SlamDynamicObjects,
                 trajectory: Optional[str],
                 time_specification: TimeSpecification,
                 source_specification: SourceSpecification,
                 calibration_id: int,
                 sequence_id: int,
                 sub_sequence_id: int,
                 settings: Optional[dict] = None):
        super().__init__(external_id, calibration_id, source_specification)
        self.vehicle_data = vehicle_data
        self.dynamic_objects = dynamic_objects
        self.trajectory = trajectory
        self.time_specification = time_specification
        self.sequence_id = sequence_id
        self.sub_sequence_id = sub_sequence_id
        self.settings = settings

    def to_dict(self):
        as_dict = super().to_dict()
        as_dict["vehicleData"] = self.vehicle_data
        as_dict["dynamicObjects"] = self.dynamic_objects.to_dict()
        as_dict["timeSpecification"] = self.time_specification.to_dict()
        as_dict["sequenceId"] = self.sequence_id
        as_dict["subSequenceId"] = self.sub_sequence_id

        if self.settings:
            as_dict["settings"] = self.settings

        if self.trajectory:
            as_dict["trajectory"] = self.trajectory

        return as_dict


class SlamFiles(RequestCall):
    def __init__(self, point_clouds: List[PointCloud], videos: Optional[List[Video]]):
        self.point_clouds = point_clouds
        self.videos = videos

    def to_dict(self):
        as_dict = dict(pointClouds=[pc.to_dict() for pc in self.point_clouds])
        if self.videos is not None:
            as_dict['videos'] = [video.to_dict() for video in self.videos]

        return as_dict


class FilesToUpload(RequestCall):
    """
    Used when retrieving upload urls from input api
    """

    def __init__(self, files: List[str]):
        self.files = files

    def to_dict(self):
        return dict(filesToUpload=self.files)


class PoseTransform(RequestCall):
    def __init__(self, timestamp: int, utc_timestamp: float, position: List[float],
                 rotation_quaternion: List[float]):
        self.timestamp = timestamp
        self.utc_timestamp = utc_timestamp
        self.position = position
        self.rotation_quaternion = rotation_quaternion

    def to_dict(self):
        return {
            "timestamp": self.timestamp,
            "utc_timestamp": self.utc_timestamp,
            "position": self.position,
            "rotation_quaternion": self.rotation_quaternion
        }


class Trajectory(RequestCall):
    def __init__(self, trajectory: List[PoseTransform]):
        self.trajectory = trajectory

    def to_dict(self):
        return [pt.to_dict() for pt in self.trajectory]
#
# Responses
#


class CalibrationNoContent(Response):
    def __init__(self, id: int, external_id: str, created: datetime):
        self.id = id
        self.external_id = external_id
        self.created = created

    @staticmethod
    def from_json(js: dict):
        return CalibrationNoContent(
            int(js["id"]), js["externalId"], ts_to_dt(js["created"])
        )

    def __repr__(self):
        return f"<CalibrationNoContent(" + \
            f"id={self.id}, " + \
            f"external_id={self.external_id}, " + \
            f"created={self.created})>"


class CalibrationWithContent(Response):
    def __init__(self, id: int, external_id: str, created: datetime,
                 calibration: Mapping[str, dict]):
        self.id = id
        self.external_id = external_id
        self.created = created
        self.calibration = calibration

    @staticmethod
    def from_json(js: dict):
        return CalibrationWithContent(int(js["id"]), js["externalId"],
                                      ts_to_dt(js["created"]), js["calibration"])

    def __repr__(self):
        return f"<CalibrationWithContent(" + \
            f"id={self.id}, " + \
            f"external_id={self.external_id}, " + \
            f"created={self.created}, " + \
            f"calibration={{...}})>"


class InputList(Response):
    def __init__(self, id: int, project_id: int, name: str, created: datetime):
        self.id = id
        self.project_id = project_id
        self.name = name
        self.created = created

    @staticmethod
    def from_json(js: dict):
        return InputList(int(js["id"]), int(js["projectId"]), js["name"],
                         ts_to_dt(js["created"]))

    def __repr__(self):
        return f"<InputList(" + \
            f"id={self.id}, " + \
            f"project_id={self.project_id}, " + \
            f"name={self.name}, " + \
            f"created={self.created})>"


class InputBatch(Response):
    def __init__(self, external_id: str, title: str, status: InputBatchStatus,
                 created: datetime, updated: datetime):
        self.external_id = external_id
        self.title = title
        self.status = status
        self.created = created
        self.updated = updated

    @staticmethod
    def from_json(js: dict):
        return InputBatch(js["externalId"], js["title"], js["status"],
                          ts_to_dt(js["created"]), ts_to_dt(js["updated"]))

    def __repr__(self):
        return f"<InputBatch(" + \
            f"external_id={self.external_id}, " + \
            f"title={self.title}, " + \
            f"status={self.status}, " + \
            f"created={self.created}, " + \
            f"updated={self.updated})>"


class Project(Response):
    def __init__(self, created: datetime, title: str, description: str,
                 status: str, external_id: str):
        self.created = created
        self.title = title
        self.description = description
        self.status = status
        self.external_id = external_id

    @staticmethod
    def from_json(js: dict):
        return Project(ts_to_dt(js["created"]), js["title"],
                       js["description"], js["status"], js["externalId"])

    def __repr__(self):
        return f"<Project(" + \
            f"created={self.created}, " + \
            f"title={self.title}, " + \
            f"description={self.description}, " + \
            f"status={self.status}, " + \
            f"external_id={self.external_id})>"


class Request(Response):
    def __init__(self, id: int, created: datetime, project_id: int, title: str, description: str,
                 input_list_id: int, input_batch_id: int, external_id: str):
        self.id = id
        self.created = created
        self.project_id = project_id
        self.title = title
        self.description = description
        self.input_list_id = input_list_id
        self.input_batch_id = input_batch_id
        self.external_id = external_id

    @staticmethod
    def from_json(js: dict):
        return Request(int(js["id"]), ts_to_dt(js["created"]), int(js["projectId"]),
                       js["title"], js["description"], int(js["inputListId"]),
                       int(js["inputBatchId"]), js["externalId"])

    def __repr__(self):
        return f"<Request(" + \
            f"id={self.id}, " + \
            f"created={self.created}, " + \
            f"project_id={self.project_id}, " + \
            f"title={self.title}, " + \
            f"description={self.description}, " + \
            f"input_list_id={self.input_list_id}, " + \
            f"input_batch_id={self.input_batch_id}, " + \
            f"external_id={self.external_id})>"


class ExportAnnotation(Response):
    def __init__(self, annotation_id: int, export_content: dict):
        self.annotation_id = annotation_id
        self.export_content = export_content

    @staticmethod
    def from_json(js: dict):
        return ExportAnnotation(int(js["annotationId"]), js["exportContent"])

    def __repr__(self):
        return f"<ExportAnnotation(" + \
            f"annotation_id={self.annotation_id}, " + \
            f"export_content={{...}})>"


class InputJob(Response):
    def __init__(self, id: int, internal_id: str, external_id: str, filename: str,
                 status: str, added: datetime, error_message: Optional[str]):
        self.id = id
        self.internal_id = internal_id
        self.external_id = external_id
        self.filename = filename
        self.status = status
        self.added = added
        self.error_message = error_message

    @staticmethod
    def from_json(js: dict):
        return InputJob(int(js["id"]), js["jobId"], js["externalId"], js["filename"],
                        str(js["status"]), ts_to_dt(js["added"]), js.get("errorMessage"))

    def __repr__(self):
        return f"<InputJob(" + \
            f"id={self.id}, " + \
            f"internal_id={self.internal_id}, " + \
            f"external_id={self.external_id}, " + \
            f"filename={self.filename}, " + \
            f"status={self.status}, " + \
            f"added={self.added}, " + \
            f"error_message={self.error_message})>"


class CreateInputJobResponse(Response):
    def __init__(self, internal_id: int):
        self.internal_id = internal_id

    @staticmethod
    def from_json(js: dict):
        return CreateInputJobResponse(js["internalId"])

    def __repr__(self):
        return f"<CreateInputJobResponse(" + \
               f"internal_id={self.internal_id})>"


class Data(Response):
    def __init__(self, id, external_id: str, source: Optional[str], created: datetime):
        self.id = id
        self.external_id = external_id
        self.source = source
        self.created = created

    @staticmethod
    def from_json(js: dict):
        return Data(
            int(js["id"]),
            str(js.get("externalId")),
            js.get("source"),
            ts_to_dt(js["created"])
        )

    def __repr__(self):
        return f"<Data(" + \
            f"id={self.id}, " + \
            f"external_id={self.external_id}, " + \
            f"source={self.source}, " + \
            f"created={self.created})>"


class Input(Response):
    def __init__(
        self,
        internal_id: str,
        external_id: str,
        batch: str,
        input_type: str,
        status: InputStatus,
        error_message: Optional[str],
    ):
        self.internal_id = internal_id
        self.external_id = external_id
        self.batch = batch
        self.input_type = input_type
        self.status = status
        self.error_message = error_message

    @staticmethod
    def from_json(js: dict):
        return Input(
            js["internalId"],
            js["externalId"],
            js["batchId"],
            js["inputType"],
            js.get("status"),
            js.get("errorMessage")
        )

    def __repr__(self):
        return f"<Input(" + \
            f"internal_id={self.internal_id}, " + \
            f"external_id={self.external_id}, " + \
            f"input_type={self.input_type}, " + \
            f"batch={self.batch}, " + \
            f"status={self.status})>"


class InvalidatedInputsResponse(Response):
    def __init__(self, invalidated_input_ids: List[int], not_found_input_ids: List[int],
                 already_invalidated_input_ids: List[int]):
        self.invalidated_input_ids = invalidated_input_ids
        self.not_found_input_ids = not_found_input_ids
        self.already_invalidated_input_ids = already_invalidated_input_ids

    @staticmethod
    def from_json(js: dict):
        return InvalidatedInputsResponse(js["invalidatedInputIds"],
                                         js["notFoundInputIds"],
                                         js["alreadyInvalidatedInputIds"])

    def __repr__(self):
        return f"<InvalidatedInputsResponse(" + \
               f"invalidated_input_ids={self.invalidated_input_ids}, " + \
               f"not_found_input_ids={self.not_found_input_ids}, " + \
               f"already_invalidated_input_ids={self.already_invalidated_input_ids})>"


class RemovedInputsResponse(Response):
    def __init__(self, removed_input_ids: List[int], not_found_input_ids: List[int],
                 already_removed_input_ids: List[int]):
        self.removed_input_ids = removed_input_ids
        self.not_found_input_ids = not_found_input_ids
        self.already_removed_input_ids = already_removed_input_ids

    @staticmethod
    def from_json(js: dict):
        return InvalidatedInputsResponse(js["removedInputIds"],
                                         js["notFoundInputIds"],
                                         js["alreadyRemovedInputIds"])

    def __repr__(self):
        return f"<RemovedInputsResponse(" + \
               f"removed_input_ids={self.removed_input_ids}, " + \
               f"not_found_input_ids={self.not_found_input_ids}, " + \
               f"already_removed_input_ids={self.already_removed_input_ids})>"


class UploadUrlsResponse(Response):
    def __init__(self, files_to_url: Dict[str, str], internal_id: int):
        self.files_to_url = files_to_url
        self.internal_id = internal_id

    @staticmethod
    def from_json(js: dict):
        return UploadUrlsResponse(js["files"], js["jobId"])

    def __repr__(self):
        return f"<UploadUrlsResponse(" + \
               f"files_to_url={self.files_to_url}, " + \
               f"internal_id={self.internal_id})>"



from dataclasses import dataclass
from typing import List, Dict, Optional
from .abstract_models import RequestCall, SensorCalibration
from .enums import CameraType


@dataclass
class RotationQuaternion(RequestCall):
    w: float
    x: float
    y: float
    z: float

    def to_dict(self) -> Dict:
        return self.__dict__


@dataclass
class Position(RequestCall):
    x: float
    y: float
    z: float

    def to_dict(self) -> Dict:
        return self.__dict__


@dataclass
class CameraMatrix(RequestCall):
    fx: float
    fy: float
    cx: float
    cy: float

    def to_dict(self) -> Dict:
        return self.__dict__


@dataclass
class DistortionCoefficients(RequestCall):
    k1: float
    k2: float
    p1: float
    p2: float
    k3: Optional[float] = None

    def to_dict(self) -> Dict:
        return {k: v for k, v in self.__dict__.items() if v is not None}

    def validate(self, camera_type: CameraType):
        if camera_type == CameraType.KANNALA:
            assert(self.k3 is None)
        else:
            assert(self.k3 is not None)


@dataclass
class UndistortionCoefficients(RequestCall):
    l1: float
    l2: float
    l3: float
    l4: float

    def to_dict(self) -> Dict:
        return self.__dict__


@dataclass
class CameraProperty(RequestCall):
    camera_type: CameraType

    def to_dict(self):
        return {
            "camera_type": self.camera_type
        }

    def get_camera_type(self) -> CameraType:
        return self.camera_type


@dataclass
class LidarCalibrationExplicit(SensorCalibration):
    position: Position
    rotation_quaternion: RotationQuaternion

    def to_dict(self) -> Dict:
        return {
            "position": self.position.to_dict(),
            "rotation_quaternion": self.rotation_quaternion.to_dict()
        }


@dataclass
class CameraCalibrationExplicit(SensorCalibration):
    position: Position
    rotation_quaternion: RotationQuaternion
    camera_matrix: CameraMatrix
    distortion_coefficients: DistortionCoefficients
    camera_properties: CameraProperty
    image_height: int
    image_width: int
    undistortion_coefficients: Optional[UndistortionCoefficients] = None

    def __post_init__(self):
        camera_type = self.camera_properties.get_camera_type()
        self.distortion_coefficients.validate(camera_type=camera_type)
        if camera_type == CameraType.KANNALA:
            assert (self.undistortion_coefficients is not None)

    def to_dict(self) -> Dict:
        base = {
            "position": self.position.to_dict(),
            "rotation_quaternion": self.rotation_quaternion.to_dict(),
            "camera_matrix": self.camera_matrix.to_dict(),
            "camera_properties": self.camera_properties.to_dict(),
            "distortion_coefficients": self.distortion_coefficients.to_dict(),
            "image_height": self.image_height,
            "image_width": self.image_width
        }
        if self.undistortion_coefficients is not None:
            base["undistortion_coefficients"] = self.undistortion_coefficients.to_dict()

        return base



"""Client for communicating with the Annotell platform."""
import logging
import mimetypes
import random
from pathlib import Path
from typing import List, Mapping, Optional, Union, Dict, BinaryIO
from uuid import uuid4 as uuid
from annotell.input_api.util import filter_none

import requests
import time
from PIL import Image
from annotell.auth.authsession import (
    FaultTolerantAuthRequestSession, DEFAULT_HOST as DEFAULT_AUTH_HOST
)

from . import input_api_model as IAM

DEFAULT_HOST = "https://input.annotell.com"

log = logging.getLogger(__name__)

RETRYABLE_STATUS_CODES = [408, 429, 500, 501, 502, 503,
                          504, 505, 506, 507, 508, 509, 510, 511, 598, 599]


class InputApiClient:
    """Creates Annotell inputs from local files."""

    def __init__(self, *,
                 auth=None,
                 host: str = DEFAULT_HOST,
                 auth_host: str = DEFAULT_AUTH_HOST,
                 client_organization_id: int = None,
                 max_upload_retry_attempts: int = 23,
                 max_upload_retry_wait_time: int = 60):
        """
        :param auth: auth credentials, see
        https://github.com/annotell/annotell-python/tree/master/annotell-auth
        :param host: override for input api url
        :param auth_host: override for authentication url
        :param client_organization_id: Overrides your users organization id.
        Only works with an Annotell user.
        :param max_upload_retry_attempts: Max number of attempts to retry uploading a file to GCS.
        :param max_upload_retry_wait_time:  Max with time before retrying an upload to GCS.
        """

        self.host = host
        self._auth_req_session = FaultTolerantAuthRequestSession(host=auth_host, auth=auth)
        self.headers = {
            "Accept-Encoding": "gzip",
            "Accept": "application/json"
        }
        self.dryrun_header = {"X-Dryrun": ""}

        self.MAX_NUM_UPLOAD_RETRIES = max_upload_retry_attempts
        self.MAX_RETRY_WAIT_TIME = max_upload_retry_wait_time  # seconds
        if client_organization_id is not None:
            self.headers["X-Organization-Id"] = str(client_organization_id)
            c_org_id = client_organization_id
            log.info(f"WARNING: You will now act as if you are part of organization: {c_org_id}. "
                     f"This will not work unless you are an Annotell user.")

    @property
    def session(self):
        return self._auth_req_session

    @staticmethod
    def _raise_on_error(resp: requests.Response) -> requests.Response:
        try:
            resp.raise_for_status()
        except requests.HTTPError as exception:
            if (exception.response is not None and exception.response.status_code == 400):
                message = ""
                try:
                    message = exception.response.json()["message"]
                except ValueError:
                    message = exception.response.text
                raise RuntimeError(message) from exception

            raise exception from None
        return resp

    def _get_upload_urls(self, files_to_upload: IAM.FilesToUpload):
        """Get upload urls to cloud storage"""
        url = f"{self.host}/v1/inputs/upload-urls"
        resp = self.session.get(url, json=files_to_upload.to_dict(), headers=self.headers)
        json_resp = self._raise_on_error(resp).json()
        return IAM.UploadUrlsResponse.from_json(json_resp)

    @staticmethod
    def _set_images_dimensions(folder: Path, images: List[IAM.Image]) -> None:
        def _is_image_missing_dimensions(img: IAM.Image):
            return img.width is None or img.height is None

        for image in images:
            if _is_image_missing_dimensions(image):
                fi = folder.joinpath(image.filename).expanduser()
                with Image.open(fi) as im:
                    width, height = im.size
                    image.height = height
                    image.width = width

    @staticmethod
    def _unwrap_enveloped_json(js: dict) -> dict:
        if js.get(IAM.ENVELOPED_JSON_TAG) is not None:
            return js[IAM.ENVELOPED_JSON_TAG]
        return js

    @staticmethod
    def _get_content_type(filename: str) -> str:
        # https://developer.mozilla.org/en-US/docs/Web/HTTP/Basics_of_HTTP/MIME_types/Complete_list_of_MIME_types
        if filename.split(".")[-1] == "csv":
            content_type = "text/csv"
        else:
            ct = mimetypes.guess_type(filename)[0]
            content_type = ct if ct is not None else 'application/octet-stream'

        return content_type

    def _get_wait_time(self, upload_attempt: int) -> int:
        """
        Calculates the wait time before attempting another file upload to GCS

        :param upload_attempt: How many attempts to upload that have been made
        :return: int: The time to wait before retrying upload
        """
        max_wait_time = pow(2, upload_attempt - 1)
        wait_time = random.random()*max_wait_time
        wait_time = wait_time if wait_time < self.MAX_RETRY_WAIT_TIME else self.MAX_RETRY_WAIT_TIME
        return wait_time

    #  Using similar retry strategy as gsutil
    #  https://cloud.google.com/storage/docs/gsutil/addlhelp/RetryHandlingStrategy
    def _upload_file(
            self, upload_url: str, file: BinaryIO, headers: Dict[str, str],
            upload_attempt: int = 1) -> None:
        """
        Upload the file to GCS, retries if the upload fails with some specific status codes.
        """
        log.info(f"Uploading file={file.name}")
        resp = self.session.put(upload_url, data=file, headers=headers)
        try:
            resp.raise_for_status()
        except requests.HTTPError as e:
            log.error(f"On upload attempt ({upload_attempt}/{self.MAX_NUM_UPLOAD_RETRIES}) to GCS "
                      f"got response:\n{resp.status_code}: {resp.content}")

            attempt_cond = upload_attempt < self.MAX_NUM_UPLOAD_RETRIES
            status_code_cond = resp.status_code in RETRYABLE_STATUS_CODES

            if attempt_cond and status_code_cond:
                wait_time = self._get_wait_time(upload_attempt)
                log.info(f"Waiting {int(wait_time)} seconds before retrying")
                time.sleep(wait_time)
                self._upload_file(upload_url, file, headers, upload_attempt + 1)
            else:
                raise e

        except Exception as e:
            raise e

    def _upload_files(self, folder: Path, url_map: Mapping[str, str]) -> None:
        """Upload all files to cloud storage"""
        for (filename, upload_url) in url_map.items():
            file_path = folder.joinpath(filename).expanduser()
            with file_path.open('rb') as file:
                content_type = self._get_content_type(filename)
                headers = {"Content-Type": content_type}
                self._upload_file(upload_url, file, headers)

    def _resolve_request_url(self,
                             resource_path: str,
                             project: Optional[str] = None,
                             batch: Optional[str] = None) -> str:
        """
        Resolves which request url to use for input based on if project and batch is specified
        """
        url = f"{self.host}/v1/inputs/"

        if project is not None:
            url += f"project/{project}/"
            if batch is not None:
                url += f"batch/{batch}/"

        url += resource_path

        return url

    def _post_input_request(self, resource_path: str,
                            input_request: dict,
                            project: Optional[str],
                            batch: Optional[str],
                            input_list_id: Optional[int],
                            dryrun: bool = False):
        if dryrun:
            headers = {**self.headers, **self.dryrun_header}
        else:
            headers = {**self.headers}

        if (input_list_id is not None):
            input_request['inputListId'] = input_list_id

        request_url = self._resolve_request_url(resource_path, project, batch)
        resp = self.session.post(request_url, json=input_request, headers=headers)
        json_resp = self._unwrap_enveloped_json(self._raise_on_error(resp).json())
        if not dryrun:
            return IAM.CreateInputJobResponse.from_json(json_resp)

    def _create_inputs_point_cloud_with_images(
            self, point_clouds_with_images: IAM.PointCloudsWithImages,
            internal_id: str,
            metadata: IAM.CalibratedSceneMetaData,
            project: Optional[str],
            batch: Optional[str],
            input_list_id: Optional[int],
            dryrun: bool = False) -> Optional[IAM.CreateInputJobResponse]:
        """Create point cloud with images"""

        js = dict(
            files=point_clouds_with_images.to_dict(),
            internalId=internal_id,
            metadata=metadata.to_dict())

        resp = self._post_input_request(
            'pointclouds-with-images',
            js,
            project=project,
            batch=batch,
            input_list_id=input_list_id,
            dryrun=dryrun
        )
        return resp

    def _create_inputs_point_clouds(self, point_clouds: IAM.PointCloudFiles,
                                    internal_id: str,
                                    metadata: IAM.SceneMetaData,
                                    project: Optional[str],
                                    batch: Optional[str],
                                    input_list_id: Optional[int],
                                    dryrun: bool = False) -> Optional[IAM.CreateInputJobResponse]:
        """Create point clouds"""

        js = dict(
            files=point_clouds.to_dict(),
            internalId=internal_id,
            metadata=metadata.to_dict())

        resp = self._post_input_request(
            'pointclouds',
            js,
            project=project,
            batch=batch,
            input_list_id=input_list_id,
            dryrun=dryrun
        )
        return resp

    def _create_images_input_job(self, images_files: IAM.ImagesFiles,
                                 metadata: IAM.SceneMetaData,
                                 internal_id: str,
                                 project: Optional[str],
                                 batch: Optional[str],
                                 input_list_id: Optional[int],
                                 dryrun: bool = False) -> Optional[IAM.CreateInputJobResponse]:
        """
        Creates an input job for an image input

        :param images_files: Contains all images, with their dimensions
        :param metadata: Contains necessary metadata in order to create and validate inputs
        :param input_list_id: input list id which the new input will be added to
        :param internal_id: When created, the input will use this internal id.
        :param dryrun: If True the files/metadata will be validated but no input job will
        be created.
        :returns CreateInputJobResponse: Class containing id of the created input job, or None
        if dryrun
        """

        create_input_job_json = dict(files=images_files.to_dict(),
                                     metadata=metadata.to_dict(),
                                     internalId=internal_id)
        resp = self._post_input_request(
            'images', create_input_job_json, project=project, batch=batch,
            input_list_id=input_list_id, dryrun=dryrun
        )
        return resp

    def create_inputs_images(
            self, folder: Path,
            images_files: IAM.ImagesFiles,
            metadata: IAM.SceneMetaData = IAM.SceneMetaData(external_id=str(uuid())),
            project: Optional[str] = None,
            batch: Optional[str] = None,
            input_list_id: Optional[int] = None,
            dryrun: bool = False) -> Optional[IAM.CreateInputJobResponse]:
        """
        Verifies the images and metadata given and then uploads images to Google Cloud Storage and
        creates an input job.
        :param folder: Absolute path to directory containing all images.
        :param images_files: List containing all images for the input.
        :param metadata: class containing metadata necessary for creating input from images.
        :param project: project to add input to
        :param batch: batch, defaults to latest open batch
        :param input_list_id: input list to add input to (alternative to project-batch)
        :param dryrun: If True the files/metadata will be validated but no input job
        will be created.
        :returns InputJobCreatedMessage: Class containing id of the created input job,
        or None if dryrun.
        """

        self._set_images_dimensions(folder, images_files.images)

        filenames = [image.filename for image in images_files.images]
        upload_url_resp = self._get_upload_urls(IAM.FilesToUpload(filenames))

        internal_id = upload_url_resp.internal_id
        self._create_images_input_job(images_files=images_files,
                                      metadata=metadata,
                                      internal_id=internal_id,
                                      project=project,
                                      batch=batch,
                                      input_list_id=input_list_id,
                                      dryrun=True)

        files_in_response = upload_url_resp.files_to_url.keys()
        assert set(filenames) == set(files_in_response)

        if not dryrun:
            self._upload_files(folder, upload_url_resp.files_to_url)
            input_job_created_message = self._create_images_input_job(images_files=images_files,
                                                                      metadata=metadata,
                                                                      internal_id=internal_id,
                                                                      project=project,
                                                                      batch=batch,
                                                                      input_list_id=input_list_id
                                                                      )
            if input_job_created_message:
                int_id = input_job_created_message.internal_id
                log.info(
                    f"Creating input for images with internal_id={int_id}")
            return input_job_created_message
        return None

    def create_inputs_point_clouds(
            self, folder: Path,
            point_clouds: IAM.PointCloudFiles,
            metadata: IAM.SceneMetaData = IAM.SceneMetaData(external_id=str(uuid())),
            project: Optional[str] = None,
            batch: Optional[str] = None,
            input_list_id: Optional[int] = None,
            dryrun: bool = False) -> Optional[IAM.CreateInputJobResponse]:
        """
        Upload files and create an input of type 'point_cloud'.

        :param folder: path to folder containing files
        :param point_clouds: list of pointclouds that constitute the input
        :param metadata: Class containing metadata necessary for point cloud
        :param project: project to add input to
        :param batch: batch, defaults to latest open batch
        :param input_list_id: input list to add input to (alternative to project-batch)
        :param dryrun: If True the files/metadata will be validated but no input job will
        be created.
        :returns CreateInputJobResponse: Class containing id of the created input job,
        or None if dryrun.

        The files are uploaded to annotell GCS and an input_job is submitted to the inputEngine.
        In order to increase annotation tool performance the supplied pointcloud-file is converted
        into potree after upload (server side). Supported fileformats for pointcloud files are
        currently .csv & .pcd (more information about formatting can be found in the readme.md).
        The job is successful once it converts the pointcloud file into potree, at which time an
        input of type 'point_cloud' is created for the designated `project` `batch` or
        `input_list_id`. If the input_job fails (cannot perform conversion) the input is not added.
        To see if conversion was successful please see the method `get_input_jobs_status`.
        """

        files_on_disk = [pc.filename for pc in point_clouds.point_clouds]

        upload_urls_response = self._get_upload_urls(IAM.FilesToUpload(files_on_disk))

        files_in_response = list(upload_urls_response.files_to_url.keys())
        assert set(files_on_disk) == set(files_in_response)

        self._create_inputs_point_clouds(point_clouds,
                                         upload_urls_response.internal_id,
                                         metadata,
                                         project=project,
                                         batch=batch,
                                         input_list_id=input_list_id,
                                         dryrun=True)
        if not dryrun:
            # self._upload_files(folder, upload_urls_response.files_to_url)

            create_input_response = self._create_inputs_point_clouds(
                point_clouds,
                upload_urls_response.internal_id,
                metadata,
                project=project,
                batch=batch,
                input_list_id=input_list_id,
            )
            if create_input_response:
                log.info(
                    f"Creating inputs for files with job_id={create_input_response.internal_id}"
                )
            return create_input_response
        return None

    def create_inputs_point_cloud_with_images(
            self, folder: Path,
            point_clouds_with_images: IAM.PointCloudsWithImages,
            metadata: IAM.CalibratedSceneMetaData,
            project: Optional[str] = None,
            batch: Optional[str] = None,
            input_list_id: Optional[int] = None,
            dryrun: bool = False) -> Optional[IAM.CreateInputJobResponse]:
        """
        Upload files and create an input of type 'point_cloud_with_image'.

        :param folder: path to folder containing files
        :param point_clouds_with_images: class containing images and pointclouds that
        constitute the input
        :param metadata: Class containing metadata necessary for point cloud with images
        :param project: project to add input to
        :param batch: batch, defaults to latest open batch
        :param input_list_id: input list to add input to (alternative to project-batch)
        :param dryrun: If True the files/metadata will be validated but no input job will be
        created.
        :returns CreateInputJobResponse: Class containing id of the created input job,
        or None if dryrun.

        The files are uploaded to annotell GCS and an input_job is submitted to the inputEngine.
        In order to increase annotation tool performance the supplied pointcloud-file is converted
        into potree after upload (server side). Supported fileformats for pointcloud files are
        currently .csv & .pcd (more information about formatting can be found in the readme.md).
        The job is successful once it converts the pointcloud file into potree, at which time an
        input of type 'point_cloud_with_image' is created for the designated `project` `batch` or
        `input_list_id`. If the input_job fails (cannot perform conversion) the input is not added.
        To see if conversion was successful please see the method `get_input_jobs_status`.
        """

        files_on_disk = [image.filename for image in point_clouds_with_images.images] + \
                        [pc.filename for pc in point_clouds_with_images.point_clouds]

        upload_urls_response = self._get_upload_urls(IAM.FilesToUpload(files_on_disk))

        files_in_response = list(upload_urls_response.files_to_url.keys())
        assert set(files_on_disk) == set(files_in_response)

        self._set_images_dimensions(folder, point_clouds_with_images.images)
        self._create_inputs_point_cloud_with_images(point_clouds_with_images,
                                                    upload_urls_response.internal_id,
                                                    metadata,
                                                    project=project,
                                                    batch=batch,
                                                    input_list_id=input_list_id,
                                                    dryrun=True)
        if not dryrun:
            self._upload_files(folder, upload_urls_response.files_to_url)

            create_input_response = self._create_inputs_point_cloud_with_images(
                point_clouds_with_images,
                upload_urls_response.internal_id,
                metadata,
                project=project,
                batch=batch,
                input_list_id=input_list_id,
            )

            if create_input_response:
                log.info(
                    f"Creating inputs for files with job_id={create_input_response.internal_id}"
                )
            return create_input_response
        return None

    def create_slam_input_job(self, slam_files: IAM.SlamFiles,
                              metadata: IAM.SlamMetaData,
                              project: Optional[str] = None,
                              batch: Optional[str] = None,
                              input_list_id: Optional[int] = None,
                              dryrun=False) -> Optional[IAM.CreateInputJobResponse]:
        """
        Creates a slam input job, then sends a message to inputEngine which will request for a
        SLAM job to be started.

        :param slam_files: class containing files necessary for SLAM.
        :param metadata: class containing metadata necessary for SLAM.
        :param project: project to add input to
        :param batch: batch, defaults to latest open batch
        :param input_list_id: input list to add input to (alternative to project-batch)
        :param dryrun: If True the files/metadata will be validated but
        no input job will be created.
        :returns CreateInputJobResponse: Class containing id of the created input job,
        or None if dryrun.
        """

        slam_json = dict(files=slam_files.to_dict(),
                         metadata=metadata.to_dict(), inputListId=input_list_id)

        resp = self._post_input_request(
            'slam',
            slam_json,
            project=project,
            batch=batch,
            input_list_id=input_list_id,
            dryrun=dryrun
        )
        return resp

    def update_completed_slam_input_job(self, pointcloud_uri: str,
                                        trajectory: IAM.Trajectory,
                                        job_id: str) -> None:
        """
        Updates an input job with data about the created SLAM, then sends a message to inputEngine
        which will create an input. The method will throw an error if the operation was
        unsuccessful.

        :param pointcloud_uri: URI pointing to a SLAM:ed pointcloud in either s3 or gs cloud
        storage.
        :param trajectory: class containing the trajectory of the SLAM:ed pointcloud.
        :param job_id: UUID for the input job.
        :returns None
        """
        url = f"{self.host}/v1/inputs/progress"
        update_json = dict(files=dict(pointClouds=pointcloud_uri),
                           metadata=dict(trajectory=trajectory.to_dict()),
                           jobId=job_id)
        resp = self.session.post(url, json=update_json, headers=self.headers)
        self._raise_on_error(resp).json()

    def update_failed_slam_input_job(self, job_id: str, message: str) -> None:
        """
        Updates an input job with an error message, then sends a message to inputEngine which will
        notify the responsible party about the failed input job. The method will throw an error if
        the operation was unsuccessful.

        :param job_id: UUID for the input job.
        :param message: String with the error message.
        :returns None
        """
        url = f"{self.host}/v1/inputs/progress"
        update_json = dict(jobId=job_id, message=message)
        resp = self.session.post(url, json=update_json, headers=self.headers)
        self._raise_on_error(resp).json()

    def get_inputs(self,
                   project: str,
                   batch: Optional[str] = None,
                   external_ids: Optional[List[str]] = None,
                   include_invalidated: bool = False) -> List[IAM.Input]:
        """
        Gets inputs for project, with option to include invalidated inputs in the list
        (by default invalidated inputs are omitted)

        :param project: Project to filter
        :param batch: Batch to filter, if omitted inputs from all batches are included.
        :param invalidated: If true, includes invalidated inputs in the response
        :return List: List of Inputs
        """

        url = f"{self.host}/v1/inputs"
        external_ids_query_string = ",".join(external_ids) if external_ids is not None else None
        params = {
            "project": project,
            "batch": batch,
            "externalIds": external_ids_query_string,
            "invalidated": include_invalidated
        }

        resp = self.session.get(url, params=filter_none(params), headers=self.headers)
        json_resp = self._unwrap_enveloped_json(self._raise_on_error(resp).json())
        return [IAM.Input.from_json(js) for js in json_resp]

    def invalidate_inputs(self,
                          input_internal_ids: List[str],
                          invalidated_reason: IAM.InvalidatedReasonInput) -> IAM.InvalidatedInputsResponse:
        """
        Invalidates inputs, and removes them from all input lists

        :param input_internal_ids: The input internal ids to invalidate
        :param invalidated_reason: An Enum describing why inputs were invalidated
        :return InvalidatedInputsResponse: Class containing what inputs were invalidated
        """
        url = f"{self.host}/v1/inputs/invalidate"
        invalidated_json = dict(inputIds=input_internal_ids, invalidatedReason=invalidated_reason)
        resp = self.session.post(url, json=invalidated_json, headers=self.headers)
        resp_json = self._unwrap_enveloped_json(self._raise_on_error(resp).json())
        return IAM.InvalidatedInputsResponse.from_json(resp_json)

    def get_projects(self) -> List[IAM.Project]:
        """
        Returns all projects connected to the users organization.

        :return List: List containing all projects connected to the user
        """
        url = f"{self.host}/v1/projects"
        resp = self.session.get(url, headers=self.headers)
        json_resp = self._unwrap_enveloped_json(self._raise_on_error(resp).json())
        return [IAM.Project.from_json(js) for js in json_resp]

    def get_project_batches(self, project: str) -> List[IAM.InputBatch]:
        """
        Returns all `batches` for the `project`.

        :param project: Project identifier
        :return List: List containing all batches
        """
        url = f"{self.host}/v1/projects/{project}/batches"
        resp = self.session.get(url, headers=self.headers)
        json_resp = self._unwrap_enveloped_json(self._raise_on_error(resp).json())
        return [IAM.InputBatch.from_json(js) for js in json_resp]

    def get_calibration_data(
        self, id: Optional[int] = None, external_id: Optional[str] = None
    ) -> List[Union[IAM.CalibrationNoContent, IAM.CalibrationWithContent]]:
        """
        Queries the Input API for either:
        * A list containing a specific calibration (of only the id is given)
        * A list of calibrations connected to an external_id (if only the external_id is given)
        * A list of calibrations connected to the users organization.
        Note that both id and external_id cannot be given at the same time.

        :param id: The id of the calibration to get
        :param external_id: The external id of the calibration(s) to get
        :return List: A list of CalibrationNoContent if an id or external id was given, or a
        list of CalibrationWithContent otherwise.
        """
        base_url = f"{self.host}/v1/inputs/calibration-data"
        if id:
            url = base_url + f"?id={id}"
        elif external_id:
            url = base_url + f"?externalId={external_id}"
        else:
            url = base_url

        resp = self.session.get(url, headers=self.headers)

        json_resp = self._raise_on_error(resp).json()
        if base_url == url:
            return [IAM.CalibrationNoContent.from_json(js) for js in json_resp]
        elif id is not None:
            return [IAM.CalibrationWithContent.from_json(json_resp)]
        else:
            return [IAM.CalibrationWithContent.from_json(js) for js in json_resp]

    def create_calibration_data(
        self, calibration_spec: IAM.CalibrationSpec
    ) -> IAM.CalibrationNoContent:
        """
        Creates a new calibration, given the CalibrationSpec
        :param calibration_spec: A CalibrationSpec instance containing everything to
        create a calibration.
        :return CalibrationNoContent: Class containing the calibration id, external id and
        time of creation.
        """
        url = f"{self.host}/v1/inputs/calibration-data"
        resp = self.session.post(url, json=calibration_spec.to_dict(), headers=self.headers)
        json_resp = self._raise_on_error(resp).json()
        return IAM.CalibrationNoContent.from_json(json_resp)

    def download_annotations(
        self, internal_ids: List[str]
    ) -> Dict[str, List[IAM.ExportAnnotation]]:
        """
        Returns the export ready annotations for each input (via the internal id).

        :param internal_ids: List with internal ids
        :return Dict: A dictionary, with each key corresponding internal_id, and value
        corresponding to a list of the ready annotations for the associated input.
        """
        url = f"{self.host}/v1/inputs/export"
        js = internal_ids
        resp = self.session.get(url, json=js, headers=self.headers)
        json_resp = self._raise_on_error(resp).json()

        for k, v in json_resp.items():
            inner_list = list()
            for kk, vv in v.items():
                inner_list.append(IAM.ExportAnnotation.from_json(vv))
            json_resp[k] = inner_list
        return json_resp

    def get_view_links(self, internal_ids: List[str]) -> Dict[str, str]:
        """
        For each given internal id returns an URL where the input can be viewed in the web app.

        :param internal_ids: List with internal ids
        :return Dict: Dictionary mapping each internal id with an URL to view the input.
        """
        view_dict = dict()
        for internal_id in internal_ids:
            view_dict[internal_id] = f"https://app.annotell.com/input-view/{internal_id}"

        return view_dict



from typing import Optional, List
from tabulate import tabulate
from .input_api_client import InputApiClient
from pprint import pprint

import click
import os

env = os.getenv("ANNOTELL_CLIENT_ORGANIZATION_ID", None)
if env:
    org_id = int(env)
    print("<" * 25,  f" Acting on behalf of organization {org_id}", 25 * ">")
else:
    org_id = None

client = InputApiClient(auth=None, client_organization_id=org_id)


def _tabulate(body, headers, title=None):
    tab = tabulate(
        body,
        headers=headers,
        tablefmt='fancy_grid',
    )
    if title:
        title_len = len(title)
        spacing = len(tab.split('\n')[0])

        roof = "=" * spacing
        floor = "=" * spacing

        left_multi = spacing // 2 - title_len // 2 - 1

        title_text_left = " " * left_multi + title
        title_text = title_text_left

        title_house = roof + '\n' + title_text + '\n' + floor
        tab = title_house + '\n' + tab
    return tab


def _get_table(sequence, headers, title=None):
    body = []
    for p in sequence:
        body.append([vars(p)[h] for h in headers])
    return _tabulate(body, headers, title)


@click.group()
def cli():
    """A CLI wrapper for Annotell utilities"""


@click.command()
@click.argument('project', nargs=1, default=None, required=False, type=str)
@click.option('--get-batches', is_flag=True)
def projects(project, get_batches):
    print()
    if project and get_batches:
        list_of_input_batches = client.get_project_batches(project)
        headers = ["external_id", "title", "status", "created", "updated"]
        tab = _get_table(list_of_input_batches, headers, "BATCHES")
        print(tab)
    elif project:
        list_of_projects = client.get_projects()
        target_project = [p for p in list_of_projects if p.external_id == project]
        headers = ["created", "title", "description", "status", "external_id"]
        tab = _get_table(target_project, headers, "PROJECTS")
        print(tab)
    else:
        list_of_projects = client.get_projects()
        headers = ["created", "title", "description", "status", "external_id"]
        tab = _get_table(list_of_projects, headers, "PROJECTS")
        print(tab)


@click.command()
@click.argument('project', nargs=1, default=None, required=True, type=str)
@click.option('--batch', nargs=1, default=None, required=False, type=str)
@click.option('--external-ids', required=False, multiple=True)
@click.option('--include-invalidated', required=False, is_flag=True)
@click.option("--view", is_flag=True)
def inputs(project, batch, external_ids, include_invalidated, view):
    print()
    if view:
        inputs = client.get_inputs(project, batch, external_ids=external_ids, include_invalidated=include_invalidated)
        view_dict = client.get_view_links([input.internal_id for input in inputs])
        body = []
        headers = ["internal_id", "view_link"]
        for internal_id, link in view_dict.items():
            body.append([
                internal_id, link
            ])
        tab = _tabulate(body, headers, title="VIEW LINKS FOR INPUTS")
        print(tab)
    else:
        inputs = client.get_inputs(project, batch, include_invalidated=include_invalidated)
        headers = ["internal_id",
                   "external_id",
                   "batch",
                   "input_type",
                   "status",
                   "error_message"]
        tab = _get_table(inputs, headers, "INPUTS")
        print(tab)


@click.command()
@click.argument('input_internal_id', nargs=1, required=True, type=str)
def view(input_internal_id):
    print()
    view_dict = client.get_view_links([input_internal_id])
    body = [[input_internal_id, view_dict[input_internal_id]]]
    headers = ["internal_id", "view_link"]
    tab = _tabulate(body, headers, title="VIEW LINK")
    print(tab)

@click.command()
@click.option('--id', nargs=1, default=None, required=False, type=int)
@click.option("--external-id", nargs=1, required=False, type=str)
@click.option('--raw', is_flag=True)
def calibration(id, external_id, raw):
    print()
    if id is not None:
        list_of_calibrations = client.get_calibration_data(id=id)
        headers = ["id", "external_id", "created"]
        tab = _get_table(list_of_calibrations, headers, "CALIBRATION")
        print(tab)
        if raw:
            print()
            [pprint(calib.calibration) for calib in list_of_calibrations]
    elif external_id is not None:
        list_of_calibrations = client.get_calibration_data(external_id=external_id)
        headers = ["id", "external_id", "created"]
        tab = _get_table(list_of_calibrations, headers, "CALIBRATION")
        print(tab)
    else:
        list_of_calibrations = client.get_calibration_data()
        headers = ["id", "external_id", "created"]
        tab = _get_table(list_of_calibrations, headers, "CALIBRATION")
        print(tab)


cli.add_command(projects)
cli.add_command(inputs)
cli.add_command(calibration)
cli.add_command(view)


def main():
    cli(prog_name="annoutil")



from setuptools import setup, find_namespace_packages
import re

URL = 'https://github.com/annotell/annotell-python'

package_name = 'annotell-input-api'

with open('README.md') as f:
    LONG_DESCRIPTION = f.read()

# resolve version by opening file. We cannot do import during install
# since the package does not yet exist
with open('annotell/input_api/__init__.py', 'r') as fd:
    match = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
                      fd.read(), re.MULTILINE)
    version = match.group(1) if match else None

if not version:
    raise RuntimeError('Cannot find version information')

# https://packaging.python.org/guides/packaging-namespace-packages/
packages = find_namespace_packages(include=['annotell.*'])

setup(
    name=package_name,
    packages=packages,
    namespace_packages=["annotell"],
    version=version,
    description='Annotell Input Api Client',
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    author='Annotell',
    author_email='Marko Cotra <marko.cotra@annotell.com>',
    license='MIT',
    url=URL,
    download_url='%s/tarball/%s' % (URL, version),
    keywords=['API', 'Annotell'],
    install_requires=[
        'annotell-auth>=1.5.0,<2',
        'annotell-cloud-storage>=0.3.0',
        'click>=7.1.1',
        'Pillow>=7.0.0',
        'requests>=2.23.0',
        'tabulate>=0.8.7',
        'python-dateutil',
        "dataclasses;python_version<'3.7'"
    ],
    python_requires='>=3.6',
    include_package_data=True,
    package_data={
        '': ['*.md', 'LICENSE'],
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3',
    ],
    scripts=[
        "bin/annoutil"
    ]
)
